/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */

/* By Richa Jain,
 * Indian Institute of Technology, Bombay.
 * June, 2001.
*/ 

/* Copyright (c) 2001 Indian Insitute of Technology, Bombay.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code and binary code must contain
 * the above copyright notice, this list of conditions and the following 
 * disclaimer.
 * 
 * 2. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: 
 * This product includes software developed at Indian Insitute of
 * Technology, Bombay. 
 *
 * 3. The name of the Institute may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission.
 * INDIAN INSTITUTE OF TECHNOLOGY, BOMBAY, MAKES NO REPRESENTATIONS 
 * CONCERNING EITHER THE MERCHANTABILITY OF THIS SOFTWARE OR THE 
 * SUITABILITY OF THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software 
 * is provided "as is" without express or implied warranty of any kind.
*/

/* A special slotted Aloha based MAC for GPRS/GSM  */ 

#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "mobilenode.h"
#include "address.h"     

// #define DEBUG
//#include <debug.h>

#include "arp.h"
#include "ll.h"
#include "rlc.h"
#include "queue.h" 
#include "mac.h"
#include "mac-gprs.h"
#include "cmu-trace.h"


/* Phy specs similar to 802.11 */
static PHY_MIB PMIB = {
	DSSS_CWMin, DSSS_CWMax, DSSS_SlotTime, DSSS_CCATime,
	DSSS_RxTxTurnaroundTime, DSSS_SIFSTime, DSSS_PreambleLength,
	DSSS_PLCPHeaderLength
};	

/* Timers */
void MacGprsTimer::start(Packet *p, double time)
{
	Scheduler &s = Scheduler::instance();
	assert(busy_ == 0);
  
	busy_ = 1;
	paused_ = 0;
	stime = s.clock();
	rtime = time;
	assert(rtime >= 0.0);
  
	s.schedule(this, p, rtime);
}

void MacGprsTimer::stop(Packet *p) 
{
	Scheduler &s = Scheduler::instance();
	assert(busy_);
  
	if(paused_ == 0)
		s.cancel((Event *)p);
  
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
}

void BackoffGprsTimer::start(double time)
{   
	Scheduler &s = Scheduler::instance();
	assert(busy_ == 0); 

	busy_ = 1;
	paused_ = 0; 
	stime = s.clock();
	rtime = time;   
	assert(rtime >= 0.0);   
	s.schedule(this, &intr, rtime);   
}	

void BackoffGprsTimer::stop(void) 
{
	Scheduler &s = Scheduler::instance();
	assert(busy_);
  
	if(paused_ == 0)
		s.cancel(&intr);
  
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
}

void BackoffGprsTimer::handle(Event *e)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
    mac->backoffHandler();
} 
	
void SlotReleaseTimer::start(double time)
{   
	Scheduler &s = Scheduler::instance();
	assert(busy_ == 0); 
	busy_ = 1;
	paused_ = 0; 
	stime = s.clock();
	rtime = time;   
	assert(rtime >= 0.0);   
	s.schedule(this, &intr, rtime);   
}	

void SlotReleaseTimer::stop(void) 
{
	Scheduler &s = Scheduler::instance();
	assert(busy_);
	if(paused_ == 0)
		s.cancel(&intr);
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
}

void SlotReleaseTimer::handle(Event *e)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
	mac->releaseHandler();
}

void WaitTimer::start(double time)
{   
	Scheduler &s = Scheduler::instance();
	assert(busy_ == 0); 
	busy_ = 1;
	paused_ = 0; 
	stime = s.clock();
	rtime = time;   
	assert(rtime >= 0.0);   
	s.schedule(this, &intr, rtime);   
}	

void WaitTimer::stop(void) 
{
	Scheduler &s = Scheduler::instance();
	assert(busy_);
	if(paused_ == 0)
		s.cancel(&intr);
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
}

void WaitTimer::handle(Event *e)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
	mac->waitHandler();
}
/* Slot timer for TDMA scheduling - the downslot */
void DownSlotGprsTimer::handle(Event *e)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->downslotHandler(e);
}

void UpSlotGprsTimer::handle(Event *e)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->upslotHandler(e);
}
/* Receive Timer */
void RxPktGprsTimer::handle(Event *e) 
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->recvHandler(e);
}

/* Send Timer */
void TxPktGprsTimer::handle(Event *e) 
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->sendHandler(e);
}

/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */
static class MacGprsClass : public TclClass {
public:
	MacGprsClass() : TclClass("Mac/Gprs") {}
	TclObject* create(int, const char*const*) {
		return (new MacGprs(&PMIB));
	}
} class_mac_gprs;

//for the static variables:
Packet * MacGprs::rxQ[MAX_NUM_FREQ][SLOTS_PER_FRAME]={NULL};
Packet * MacGprs::txQ[MAX_NUM_FREQ][SLOTS_PER_FRAME]={NULL}; 
struct data_at_bs MacGprs::vlr_ = {0};

int MacGprs:: tx_rate = {0};
int MacGprs:: max_num_ms_ = {0};
int MacGprs:: max_num_freq_ = {0};
int MacGprs:: slot_packet_len_ = {0};
int MacGprs:: gprs_slots_per_frame_ = {0};

double MacGprs::slot_time_ = 0;
double MacGprs::start_time_ = 0;

int MacGprs::chan0_0 = {0};
int MacGprs::coll_count = {0}; 

int MacGprs::active_node_ = {0};

int MacGprs::verbose_ = {0};

//static variables for rlc_error
int MacGprs::rlc_error_  ={0};
int MacGprs::error_rate_ ={0};
int MacGprs::drop_gap_   ={0};
int MacGprs::drop_counter_ ={0};
int MacGprs::next_drop_  ={0};

//constructor

MacGprs::MacGprs(PHY_MIB* p) : Mac(), mhDownSlot_(this), 
	mhUpSlot_(this), mhTxPkt_(this), mhRxPkt_(this), 
    mhRel_(this), mhBackoff_(this),mhwait_(this)
{
	int i,j;

	phymib_ = p;

	bind("slot_packet_len_", &slot_packet_len_);
	bind("max_num_ms_", &max_num_ms_);
	bind("max_num_freq_", &max_num_freq_);
	bind("gprs_", &gprs_);
	bind("verbose_", &verbose_);
	bind("gprs_slots_per_frame_", &gprs_slots_per_frame_); 

	bind("rlc_error_", &rlc_error_);
	bind("error_rate_", &error_rate_);
	
	slot_time_= 0.000577 ;//gprs slot is of 576.9 microseconds
	tx_rate = (int) ((8*slot_packet_len_) / slot_time_) ;
		
	//printf("slot_time=%f, gprs_=%d  tx_rate=%d \n", slot_time_, gprs_, tx_rate);
	
	active_node_++;  
	
	//max_num_ms shud be less that MAX_NUM_MS
	if (max_num_ms_ > MAX_NUM_MS) {
		if (verbose_==1)
		   printf("Too many nodes taking part. shud be less than %d \n",  MAX_NUM_MS);
		exit (-1);
	}	
	if (active_node_ > max_num_ms_) {
		if (verbose_==1)
		   printf("Too many nodes taking part in the simulations, aborting...\n");
		exit(-1);
	} 
    
	radio_active_ = 0;
	wait_for_res=0;
	
	first_round_=1;
	down_slot_=0;
	up_slot_=0;
	
	for ( j=0 ; j< SLOTS_PER_FRAME ; j++) {
		pktTx[j] =NULL;
		pktRx[j] =NULL;
		tx_chan[j] = -10;
		rx_chan[j] = -10; 
		//-10 => no channel
	}
	tx_chan[0]=0;   // freq 0, slot 0 reserved. for broadcasts/isgnalling 
	rx_chan[0]=0;	// in either  direction. 

	txQ[0][0]=0;
	temp_reply=0;
	
	// initialising data_at_bs
	for ( j=0; j<max_num_ms_ ; j++) {  
		vlr_.hier_addr_[j]=-10;
	} 
	for ( j=0; j<SLOTS_PER_FRAME ; j++) {  
		 for (i=0; i<max_num_freq_ ; i++) {
			 //this freq/slot hasnt been alloted to any MS yet
	 		 vlr_.up_table[i][j]=-10;  
	 		 vlr_.down_table[i][j]=-10;
			 
		 }
	}


	if (rlc_error_==1 && index_==0) {
		drop_gap_ = Random::integer(error_rate_);
		next_drop_ = drop_gap_;
		if (verbose_==1) 
			printf("<%d> %f  first drop will be at %d \n", index_, NOW, next_drop_ );
    }
			
	if (verbose_==1)
 		printf("<%d> %f  initialised...; \n",index_, NOW);
	
	intr1_ = intr_ ; //reqd for the slot timers
	intr = intr_ ; 
	
	mhDownSlot_.start((Packet *) (& intr_), 0);  
}

// similar to 802.11
int MacGprs::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
	}
	return Mac::command(argc, argv);
}


/* ======================================================================
   Debugging Routines
   ====================================================================== */
void MacGprs::trace_pkt(Packet *p) 
{
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_gprs* dh = HDR_MAC_GPRS(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void MacGprs::dump(char *fname)
{
	fprintf(stderr, "\n%s --- (INDEX: %d, time: %2.9f)\n", fname, 
		index_, Scheduler::instance().clock());
	
	fprintf(stderr, "\tpktTx_: %x, pktRx_: %x, callback: %x\n", 
		(int) pktTx_, (int) pktRx_, (int) callback_);
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
int MacGprs::hdr_dst(char* hdr, int dst )
{
	struct hdr_mac_gprs *dh = (struct hdr_mac_gprs*) hdr;
	if(dst > -2)
		STORE4BYTE(&dst, (dh->dh_da));
	return ETHER_ADDR(dh->dh_da);
}

int MacGprs::hdr_src(char* hdr, int src )
{
	struct hdr_mac_gprs *dh = (struct hdr_mac_gprs*) hdr;
	if(src > -2)
		STORE4BYTE(&src, (dh->dh_sa));
  
	return ETHER_ADDR(dh->dh_sa);
}

int MacGprs::hdr_type(char* hdr, u_int16_t type) 
{
	struct hdr_mac_gprs *dh = (struct hdr_mac_gprs*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}


/*double DATA_Time(int len) { 
	double 
	return (8*(len) / tx_rate);
}   */   

double MacGprs::TX_Time(Packet *p) {
    
	int len=	(HDR_CMN(p))->size();
	double t = (double)(len*8)/tx_rate;
//	printf("< >, %f len=%d tx_time = %f \n", NOW,len, t);
	if(t < 0.0) {
	    drop(p, "XXX");
        exit(1);
    } 
    return t;
}   
		
/* Resource allocation:*/
void MacGprs::slot_allot( int  ms_, int &freq, int &slot)
{ // Allocation is sequential - first empty slot is alloted.
	
	int i,j;
	start_time_ = NOW;

		if (gprs_==1) { //ie its a gprs node
			//fnd the first free slot/freq
			for ( i=0; i<max_num_freq_ ; i++) {  
			    for ( j=1; j<(gprs_slots_per_frame_+1) ; j++){
					//slot0 resv for signalling on ALL freqs
					if (vlr_.up_table[i][j] == -10) {
						//ie this slot's not been alloted to anyone 
						break;
					}	
				}	
				if (j != gprs_slots_per_frame_+1 ) {
					break;
				}
			}
			if (i==max_num_freq_ && j==gprs_slots_per_frame_+1) {       		
					fprintf(stderr," No GPRS freq/slot free..... cant allot.  \n");
				return;
			}	
		} 
		else { // it's a gsm node
			for ( i=0; i<max_num_freq_ ; i++) { 
			    for ( j=gprs_slots_per_frame_+1 ; j< SLOTS_PER_FRAME; j++) {
					if (vlr_.up_table[i][j] == -10) {
						break;
					}	
				}	
				if (j != SLOTS_PER_FRAME ) {
					break;
				}
			}
			if ( i==max_num_freq_ && j==SLOTS_PER_FRAME) {       		
					fprintf(stderr," No GSM freq/slot free..... cant allot.  \n");
				return;
			}	
		} 
			
	 	// => i, j are the first free freq-slot pair.
		// reserve them for the requesting Mobile 
		vlr_.up_table[i][j] = ms_ ;
		vlr_.down_table[i][j] = ms_;
		freq=i;
		slot=j;

	//showing the current reservations	
	for (i=0; i < max_num_freq_; i++) {
	        for (j=0; j < SLOTS_PER_FRAME; j++) {          
				if (verbose_==1)
				printf ("vlr_.up_table [%d][%d] = %d \n", i,j, vlr_.up_table [i][j]  );
			}
	}			
	return;
}

/* To handle incoming packet. */
void MacGprs::recv(Packet* p, Handler* h) 
{
	MobileNode * node_ = (MobileNode*)(netif_->node());

	if (node_->base_stn() != node_->address()) {//node !=BS ie its an MS 
		ms_recv(p,h);
    }
		
	else if (node_->base_stn() == node_->address()) {//node is its own BS ie its aBS 
		bs_recv(p,h);
	}

}	
		
void MacGprs::ms_recv(Packet* p, Handler* h)
{	
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_gprs *dh = HDR_MAC_GPRS(p);
	int dst = ETHER_ADDR(dh->dh_da);   
	int src = ETHER_ADDR(dh->dh_sa);   
	
	int i,j,len; 	
	
	if (ch->direction() == hdr_cmn::UP) {
	// ie MS rx pkt from phy on a Downlink slot
			
		if ( src == 0  && (dst == index_ || ((u_int32_t)dst == MAC_BROADCAST) ) ){
			if ( rx_chan[down_slot_-1] == ch->chan() )   {
				if (verbose_==1)
				printf("<%d> %f ms_recv pkt 4 me at phy. mac_src=%d mac-dst=%d, next_hop=%s,\n			  uid= %d. ptype=%s. \n",
				index_, NOW, src, dst, Address::instance().print_nodeaddr (ch->next_hop_) ,       
				ch->uid(), packet_info.name(ch->ptype()) );
	
				if (mhRel_.busy() !=0 /*&& (ch->ptype()!=PT_ACK) */
					/*&&  (dh->dh_fc.fc_type != MAC_Type_Control)*/ ) {
				
					mhRel_.stop();
					if (verbose_==1)
					   printf("<%d> %f  pkt goin UP. release timer stopped \n", index_, NOW);
 					
					if (q_ !=NULL && gprs_ ==1){
						len= q_->length();	//restart timer if length =0
						if ( len==0 && (down_slot_ != 1) && (rx_chan[down_slot_-1] !=-10) ) {
							slot_release_time_ = (4*SLOTS_PER_FRAME) *slot_time_;
							mhRel_.start(slot_release_time_);
							if (verbose_==1)
							    printf("<%d> %f  release timer started: 1 \n",index_, NOW);
   						}
					}
				}	
				
				pktRx[down_slot_-1] = p;
				radioSwitch(ON); 
				rx_from_phy ( pktRx[down_slot_-1] );  
			}	
		}	
		return;
	}

   if (verbose_==1)
	 printf("<%d> %f ms_recv pkt at IFQ. mac_src=%d mac-dst=%d, next_hop=%s,
			 uid= %d. ptype=%s. rlc_seqno=%d, size=%d\n",
			index_, NOW, src, dst, Address::instance().print_nodeaddr (ch->next_hop_) ,       
			ch->uid(), packet_info.name(ch->ptype()), HDR_RLC(p)->seqno(), ch->size() );

	callback_ = h;
	if	( (u_int32_t)dst == MAC_BROADCAST && ch->ptype() != PT_ARP ){ 
		j = 0;
   		if (verbose_==1)
		   printf("<%d> %f Ms: pkt to broadcast dst=%d \n", NOW,dst);
	} 
	else {
		// find upslot/freq to tx on
		for ( j=1; j< SLOTS_PER_FRAME ; j++) {
			if ( tx_chan[j] != -10 ) { //ie MS has been allotted SOME upchan
				break;
			 }
		 }
		if (j==SLOTS_PER_FRAME) { // => I dont have a channel so far - ask for one.
			if (mhwait_.busy() !=0 ) {// wait timer busy - waiting for res reply..
   				if (verbose_==1)
				   printf("<%d> %f I am waiting for res_reply \n", index_, NOW);
				return; 
			} 
   			if (verbose_==1)
			   printf("<%d> %f I dont have a chan yet - will send a resource request.\n", index_, NOW);
			send_res_request();
			p_temp=p;// store the pkt to tx later.....
			return;
		}	
	}	 
 	//if we have multiple ARP requests: drop 
	if ( ch->ptype()==PT_ARP )  {
		 Packet::free(p) ;
         if (verbose_==1)
		    printf("<%d> %f I got multiple ARP requests. Dropped duplicates.\n",  index_, NOW); 
		 return;
	}	 
		
	ch->chan() = tx_chan[j]; //stamp pkt hdr with chan num
	pktTx[j] = p; // schedule pkt to be txd in upslot j
		
   if (verbose_==1)
 	  printf("<%d> %f MS: pkt 2b txd in [Uslot %d],[chan %d]. ptype=%s uid=%d\n, 
			  rlc_seqno=%d size=%d\n",
			index_, NOW, j, tx_chan[j], packet_info.name(ch->ptype()), 
			ch->uid(), HDR_RLC(p)->seqno(),ch->size() );
	
	rx_from_ll (pktTx[j] ); // add MAC hdr onto pkt
	
	
	if (mhRel_.busy() !=0 /*&&  (dh->dh_fc.fc_type != MAC_Type_Control)*/ ) {
		mhRel_.stop();
   		if (verbose_==1)
		    printf("<%d> %f  pkt goin DOWN. release timer stopped \n", index_, NOW);
   	}
	return;
}
	
		
void MacGprs::bs_recv(Packet* p, Handler* h) 
{	int i,j; 	
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_gprs *dh = HDR_MAC_GPRS(p);
	int dst = ETHER_ADDR(dh->dh_da);   
	int src = ETHER_ADDR(dh->dh_sa);   
     
	if (ch->direction() == hdr_cmn::UP) {
	// BS rxs pkt from phy at start of upslot.
    //error model: only on uplink. if u want both uplink and downlink
	//to have errors, just put these 11 lines above the if(..direction..)
	
	if (rlc_error_==1 && index_==0) {// pkts are marked  only at BS. generally
		drop_counter_ ++;
		if (drop_counter_ == next_drop_) { 
		    ch->error() = 1; //mark the pkt as erroneous
			drop_gap_ = Random::integer(error_rate_);//find next pkt to drop
			next_drop_ = next_drop_ + drop_gap_;
		    if (verbose_==1)
				printf("<%d> %f rlc pkt error. pkt %d being marked. next_drop=%d\n", 
				 index_, NOW, drop_counter_,next_drop_);
         }
    }

		if (((u_int32_t)dst == MAC_BROADCAST) || (dst == index_) ) {
			
   			if (verbose_==1)
			    printf("<%d> %f bs_recv pkt 4 me at phy, mac_src=%d mac-dst=%d, next_hop=%s,\n		  		   uid= %d. ptype=%s. upslot=%d rlc_seqno=%d \n",
			index_, NOW, src, dst, Address::instance().print_nodeaddr (ch->next_hop_) ,       
			ch->uid(), packet_info.name(ch->ptype()), (up_slot_ -1), HDR_RLC(p)->seqno()  );
			
			i = ch->chan();
			rxQ[i][up_slot_-1] = p;
			radioSwitch(ON);
			rx_from_phy ( rxQ[i][up_slot_-1] );  
		}
		return;
	}
		
	// ie for down slot: pkt rxd from IFQ. prepare to tx
	callback_ = h;

   if (verbose_==1)
	printf("<%d> %f bs_recv pkt at IFQ, mac_src=%d mac-dst=%d, next_hop=%s,\n		  uid= %d. ptype=%s. \n",
		index_, NOW, src, dst, Address::instance().print_nodeaddr (ch->next_hop_) ,       
		ch->uid(), packet_info.name(ch->ptype()) );
	
	if	( (u_int32_t) dst == MAC_BROADCAST /*|| ch->ptype()==PT_ARP*/ ){
   		if (verbose_==1)
		   printf("<%d> %f BS. have to broadcast pkt \n", index_, NOW);
	 	// downchan[0], downslot[0] reserved for broadcast/signalling 
		ch->chan()=0 ; //stamp pkt hdr
		txQ[0][0] = p ;
		rx_from_ll(txQ[0][0]);
		return;
	}
	else {
		//find the first freq/slot combo allotted for this mac-dst 
		for (i=0; i< max_num_freq_ ;i++){
			// leave out i=0 coz reserved for Broadcast/signalling
			for (j=1 ; j< SLOTS_PER_FRAME; j++ ) {
				if ( dst == vlr_.down_table[i][j]) {
					break;
				}
			}
			if (j!= SLOTS_PER_FRAME) {
   				if (verbose_==1)
			//	  printf("<%d> %f I can tx in Downslot=%d, on freq=%d\n",index_, NOW, j, i);
				break;	
			}
		}	
		if ( i==max_num_freq_ && j==SLOTS_PER_FRAME) { //eiteher dst is a fixed node
			//or it hasnt got a slot yet. 
			if (mhwait_.busy() !=0 ) {// res reply ready.. but hasnt been txed yet.....
   				if (verbose_==1)
				  printf("<%d> %f  res_reply not yet recv by MS\n", index_, NOW);
				 //return; 
			} 
			else {
   				if (verbose_==1)
				   printf("<%d> %f have to allot slots to dst=%d \n", index_, NOW, dst);	
				slot_allot( dst, i, j); 
				send_res_reply(dst,i,j);
			}	
			dh->dh_wait = 1;
   			if (verbose_==1)
		        printf("<%d> %f bs_recv dh_wait set for txQ[%d][%d]=%d \n", index_, NOW, i, j,dh->dh_wait );	
			//return;
		}
		
		ch->chan()=i ; //stamp pkt hdr
		txQ[i][j] = p ;
			
   		if (verbose_==1)
	     	printf("<%d> %f BS: pkt 2b txd in [Dslot %d],[chan %d]. ptype=%s uid=%d \n",
			index_, NOW, j, i, packet_info.name(ch->ptype()), ch->uid() );

		rx_from_ll(txQ[i][j]);
		return;
	}	
}


void MacGprs::rx_from_phy(Packet* p) 
{
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_gprs* dh = HDR_MAC_GPRS(p);      
	
	//check for collision at BS (check only at BS!) if this is a Random Access Channel
	if ( ch->chan()==0 && up_slot_ ==1 && index_==0) {
	// this is why u have t omake sure BS is index_==0!!!
		switch (chan0_0) {
			case IDLE:
   				 if (verbose_==1)
				  printf("<%d> %f uh oh!! chan0_0 idle - though a request/ARP was txd!!\n", 
						index_, NOW );
				return;
				break;
			case BUSY:
				 // recive normally
				 break;
			case COLL :
				 // if chan0_0 != 0 or 1 => many MS txing = collision
				 Packet::free(p) ; //drop the pkt
				 coll_count --;
   				 if (verbose_==1)
				    printf("<%d> %f chan0_0  collision. pkt dropped. coll_count=%d  \n" ,
						 index_, NOW, coll_count);
				 break;
		}
		if  (chan0_0 == BUSY) {
			double rtime = TX_Time(p);
			
			if (rtime > slot_time_) rtime= slot_time_ ;// hack-just in case.
			
   		    if (verbose_==1)
			   printf("<%d> %f rtime = %f \n", index_, NOW, rtime);
			assert(rtime >= 0);  
			mhRxPkt_.start(p, rtime); 
		//	mhRxPkt_.start(p, slot_time_); 
		
			chan0_0 = IDLE;
   		    if (verbose_==1)
			   printf("<%d> %f recv timr started for chan0_0, chan made IDLE \n" , index_, NOW);   
		}
	   	
		if  (chan0_0 == COLL && coll_count==0) {
			chan0_0 = IDLE;
   		    if (verbose_==1)
			   printf("<%d> %f chan0_0 made IDLE. all coll sensed.\n" , index_, NOW);   
		}
			
		return;
	}
		
	//jsut in case ..... ;) 
	if (ch->error() == 1) {
	//	Packet::free(p) ;
   	    drop(p);
		if (verbose_==1)
		   printf("<%d> %f pkt dropped. ch->error()=1 \n", index_, NOW);
		return;
	}
	
		double rtime = TX_Time(p);
		//if pkt's too large - tx/rx only for one slot interval
		if (rtime > slot_time_) rtime= slot_time_ ;
		 
   		if (verbose_==1)
		   printf("<%d> %f rtime=%f \n", index_, NOW, rtime);
		assert(rtime >= 0);  
		mhRxPkt_.start(p, rtime); 
		//mhRxPkt_.start(p, slot_time_); 
}

// Forward data to RLC/LL after pkt rxd from PHY ie when RxPktTimer times out.
// Adjust the MAC packet size: strip off the mac header for data
// and fwd to ll.
// else check which control type and deal accordingly
// NB: MAC control messgs are terminated here only and not passed
// to the RLC/LL
void MacGprs::fwd_DATA_to_LL(Packet *p)
{	int i,j, freq, slot;
	float temp;
	double backoff_time;
	
	struct hdr_cmn *ch2;
	struct hdr_mac_gprs *dh2;
	
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_gprs* dh = HDR_MAC_GPRS(p);
 	int src_ = ETHER_ADDR(dh->dh_sa); //the ode tha sent the res_request  
	
	switch ( dh->dh_fc.fc_type) {
		
		case MAC_Type_Data:				
			ch->size() -= ETHER_HDR_LEN;
			ch->num_forwards() += 1;
   		    if (verbose_==1)
			   printf("<%d> %f  fwding_data_to_LL \n", index_, NOW);
			uptarget_->recv(p, (Handler*) 0);
			break;
		
		case MAC_Type_Control:
			switch (dh->dh_fc.fc_subtype) {
				
				case MAC_Subtype_res_request: //happens at BS
					//look for next free slot in vlr and assign it 
					 slot_allot (src_, freq, slot);
   					 if (verbose_==1)
					    printf ("<%d> %f res request recvd. freq%d, slot%d reserved for <%d>.\n",
							 index_, NOW, freq, slot, src_);
				
					 send_res_reply (src_,freq, slot);
					 //	break;
					 return;
					
				case MAC_Subtype_res_reply: //happens at MS
				
					mhwait_.stop(); //stop the RA wait timer.
					// set tx/rx_chan;
					slot = dh->dh_slot;
					freq = dh->dh_freq;
					tx_chan[slot]=freq;
					rx_chan[slot]=freq; // tx and rx r symmetric
   					if (verbose_==1)
					   printf ("<%d> %f res reply recvd. freq=%d slot=%d\n", index_, NOW, freq, slot);

					// send the old pkt waiting at BS
					if ( txQ[freq][slot] ==NULL) {
					//	printf (" BS: no pkt waiting ...txq[%d][%d]  NULL \n", freq, slot);
					}
					else {
						dh2 = HDR_MAC_GPRS(txQ[freq][slot]);
						dh2->dh_wait=0;
   						if (verbose_==1)
						    printf("<%d> %f BS: OLD  pkt waiting. dh wait corrected \n", index_, NOW );
					}
					// if there's a packet buffered at the MS, send it
					if ( p_temp!= NULL) {
						ch2 = HDR_CMN(p_temp);
						ch2->chan() = freq;   //stamp pkt hdr with chan num
						pktTx[slot] = p_temp; // schedule pkt to be txd in upslot j
						p_temp= NULL;         //clear p_temp.
   						if (verbose_==1)
						    printf("<%d> %f MS: got chan. OLD pkt 2b txd in [Uslot %d],[chan %d]. p type=%s uid=%d\n",
								index_, NOW, slot, freq,packet_info.name(ch2->ptype()), ch2->uid() );
	
						rx_from_ll ( pktTx[slot] ); // add MAC hdr onto pkt
					}
					
					//unblock IFQ
				    q_->unblock();	
					if (verbose_==1) printf("<%d> %f IFQ unblocked \n", index_, NOW);
					
					return;
				    //	break;
					
				case MAC_Subtype_tx_end: //happens at BS
					//find MS entry in vlr and clear
					for (i=0; i< max_num_freq_ ;i++){
						for (j=1 ; j< gprs_slots_per_frame_; j++ ) {
							if ( src_ == vlr_.down_table[i][j]) {
   								if (verbose_==1)
								    printf("vlr entry for %d, slot%d freq%d, deleted.\n", src_,j, i );
								vlr_.down_table[i][j] = -10;
								vlr_.up_table[i][j] = -10;
								break;
							}
						}
						if ( j!= gprs_slots_per_frame_) {
							break;	
						}
					}	
					return;
				//	break;
			}	
		default:
			break;
			//nothing.. the only MAC types are data and control.
	}
}

// creates a resource_request message and schedules it to be sent
// in the next upslot 0
void MacGprs::send_res_request(){ // sent by an MS to its  BS
	
	int dst;
	Packet *p1 = Packet::alloc();
	
	struct hdr_cmn* ch1= HDR_CMN(p1);
	struct hdr_mac_gprs* dh1 = HDR_MAC_GPRS(p1);
	
	MobileNode * node_ = (MobileNode*)(netif_->node());
	
	dst=0; //hack since index_==0 => BS
    // base_stn() doesnt work here. since it hasnt yet got its 
    // ARP reply and doesnt know the MAC address of its base station!!!
	
	this->hdr_src((char*)dh1, this->addr());
	this->hdr_dst((char*)dh1, dst);

	ch1->uid() = 0;
	ch1->direction() = hdr_cmn::DOWN;
	ch1->ptype() = PT_MAC;
    //ch1->size() = slot_packet_len_;
    ch1->size() = 20; //arbitly set at 20.
    ch1->iface() = -2;
	ch1->error() = 0;  
	
	dh1->dh_fc.fc_type       = MAC_Type_Control;
	dh1->dh_fc.fc_subtype    = MAC_Subtype_res_request;
	
	ch1->chan() = 0; // send on upslot 0, freq 0 ie random access chan
	pktTx[0] = p1;


    //block the IFQ from sending down further pkts till a res_reply is
    //recvd
    q_->block();
	
    if (verbose_==1)
	    printf("<%d> %f res_request created. dst= %d. IFQ blocked.\n", 
			index_, NOW, /*node_->base_stn() */ dst);
	rx_from_ll (pktTx[0]);
}

//create a resource reply message and schedule it to be sent in
//downlsot 0
void MacGprs::send_res_reply(int dst, int freq, int slot)
{	//dst= the node that sent the res_request...
	//or the node to which a res_reply has to be sent.
	Packet *p1 = Packet::alloc();
	struct hdr_cmn* ch1= HDR_CMN(p1);
	struct hdr_mac_gprs* dh1 = HDR_MAC_GPRS(p1);
	
	this->hdr_src((char*)dh1, this->addr());
	this->hdr_dst((char*)dh1, dst);
	
	ch1->uid() = 0;
	ch1->ptype() = PT_MAC;
    //ch1->size() = slot_packet_len_;
    ch1->size() = 20; //arbitly set at 20.
    ch1->iface() = -2;
	ch1->error() = 0;  
	ch1->direction() = hdr_cmn::DOWN;
    
	if (verbose_==1)
	   printf("<%d> %f res_reply created. freq=%d, slot=%d booked for dst=%d\n",
		index_, NOW, freq, slot,dst);
			
	dh1->dh_fc.fc_type     = MAC_Type_Control;
	dh1->dh_fc.fc_subtype  = MAC_Subtype_res_reply;
	dh1->dh_freq		   = freq;	
	dh1->dh_slot		   = slot;	
	ch1->chan() = 0; // send on upslot 0, freq 0 ie random access chan

	if (txQ[0][0]==0){
		txQ[0][0] = p1;
		rx_from_ll (txQ[0][0]);
	}
	else {
		 // if I already have an as-yet-untransmitted res_reply and
		 // another comes along - store the new arrival and tx it on
		 // the next TDMA frame.
         temp_reply=p1;
		 rx_from_ll(temp_reply);
	}
//	{
	//	printf("<%d> %f i have another res_reply to send, cant send yours.\n",  index_, NOW);
	//	vlr_.down_table[freq][slot]==-10;
	//	vlr_.up_table[freq][slot]==-10;
	//	 printf("<%d> %f vlr_entry %d, %d cleared\n",index_, NOW, freq, slot);
//    }   
		
}

// Prepare pkt to be sent onto PHY. ie add headers. Is actually
// rx from IFQ! 
void MacGprs::rx_from_ll(Packet* p) 
{
	u_int32_t dst, src, size;
	struct hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac_gprs* dh = HDR_MAC_GPRS(p);

    if (verbose_==1)
	   printf("<%d> %f rxing_from_ll, size=%d\n", index_, NOW,ch->size());
	
	// Update the MAC header, same as 802.11 
	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	
	if (dh->dh_fc.fc_type  != MAC_Type_Control){
		ch->size() += ETHER_HDR_LEN;
		dh->dh_fc.fc_type       = MAC_Type_Data;
		dh->dh_fc.fc_subtype    = MAC_Subtype_data;
	} 
	
	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;

	if ( dh->dh_wait != 1) dh->dh_wait = 0; 
	
	if((u_int32_t)ETHER_ADDR(dh->dh_da) != MAC_BROADCAST)
		dh->dh_duration = DATA_DURATION;
	else
		dh->dh_duration = 0;
}

// Actually TX  the packet. 
void MacGprs::tx_onto_PHY(Packet *p) 
{
	double stime;
	double temp;
	
	struct hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac_gprs* dh  = HDR_MAC_GPRS(p);
	
	
	//check if any other MS is txing too... if this is the Random Access Channel
 	if ( ch->chan()==0 && up_slot_ ==0 && index_!=0) {
		//have to amke sure it's not a BS - coz BS may tx
		//on chan0 in downslot3
		switch (chan0_0) {
			case IDLE:
				chan0_0 = BUSY;
				if (verbose_==1)
				    printf("<%d> %f chan0_0 made BUSY \n", index_, NOW);
				break;
			case BUSY:
				coll_count = 2;
				chan0_0 = COLL;
				if (verbose_==1)
				    printf("<%d> %f chan0_0=%0x coll_count=%d \n", index_, NOW,chan0_0, coll_count);
				break;
			case COLL:
				coll_count ++;
				if (verbose_==1)
				   printf("<%d> %f chan0_0 => COLL coll_count=%d \n", index_, NOW, coll_count);
				ch->error() = 1; //mark the pkt as erroneous
				break;
		}
	}

	if (dh->dh_fc.fc_subtype == MAC_Subtype_res_request)  {
		//sent only by ms in Upslot 0. MS waits for 7 TS
		temp= (double)((SLOTS_PER_FRAME-1)*slot_time_);
		mhwait_.start(temp);
		if (verbose_==1)
		   printf("<%d> %f  wait timer started \n", index_, NOW);
	}

	stime =	TX_Time(p);   
	if (stime > slot_time_) stime= slot_time_ ;// just to make sure.
	assert(stime >= 0);  
	if (verbose_==1)
		printf("<%d> %f txing_onto_PHY, sendtime=%f, size=%d\n", index_, NOW, stime, ch->size());
	mhTxPkt_.start(p->copy(), stime);
	//mhTxPkt_.start(p->copy(), slot_time_);
	downtarget_->recv(p, this);        
}

// Turn on / off the radio
void MacGprs::radioSwitch(int i) 
{
	radio_active_ = i;
	EnergyModel *em = netif_->node()->energy_model();
	if (i == ON) {
		if (em && em->sleep()){
			em->set_node_sleep(0);
		}
		if (verbose_==1)
		    printf("<%d> %f radio switched ON\n", index_, NOW);	
		return;
	}

	if (i == OFF) {
		if (em && !em->sleep()) {
			em->set_node_sleep(1);
		};
	    if (verbose_==1)
		   printf("<%d> %f radio switched OFF\n", index_, NOW);
		return;
	}
}

		 
//============================================================
//		 Timers' handlers 
//============================================================

/* Slot Timers:*/

void MacGprs::downslotHandler(Event *e) 
{
    int i, len;
	
 	struct hdr_mac_gprs *dh;
	MobileNode * node_ = (MobileNode*)(netif_->node());
	
	if (( first_round_ ==1) & (down_slot_==0)) {

		vlr_.hier_addr_[index_] =  Address::instance().get_nodeaddr(node_->address());
		// set info in data_at_bs - since we cdnlt do it in the
		// constructor as there wasnt yet a pointer to netif_........
		if (verbose_==1)
 		  printf(" hier_Add[%d]=%d ; grps=%d \n", index_, vlr_.hier_addr_[index_], gprs_);   
	}	
		
	// Restart timer for next slot.
	mhDownSlot_.start((Packet *)e, slot_time_);
	if (index_==0 && verbose_==1)
	    printf("<%d> %f  down slot %d starts now. \n", index_, NOW, down_slot_);

	//start Upslottimer after 3 downslots to ensure skew
	if (( first_round_ ==1) & (down_slot_==3)) {
		first_round_ = 0;
	    mhUpSlot_.start((Packet *) (& intr1_), 0);  
	}	
	
	if (node_->base_stn() == node_->address()) {//node is a BS 

		for (i=0; i<max_num_freq_ ; i ++) {
			if (down_slot_==0 && txQ[i][down_slot_]==0) {
				if (temp_reply != 0) {
					radioSwitch(ON);
					tx_onto_PHY (temp_reply);
					temp_reply=0;
			     }
			}	
			// check if any chan has to tx in this slot.
			if  (txQ[i][down_slot_] !=NULL) {
				dh = HDR_MAC_GPRS(txQ[i][down_slot_]);   
				if (dh->dh_wait !=1) {
					radioSwitch(ON);
					tx_onto_PHY ( txQ[i][down_slot_] );
					txQ[i][down_slot_] = 0;
				}	
			}
		}	
	}

	if (node_->base_stn() != node_->address()) {//node is an MS 

		/* I do nothing here! all taken care of in ms_recv() */

	}	
	
	down_slot_ ++; 
	if (down_slot_ == SLOTS_PER_FRAME	) {
		down_slot_ = 0; //ie wrap around.
	}
	return;
				
}	


void MacGprs::upslotHandler(Event *e) 
{
	int i,len;
	MobileNode * node_ = (MobileNode*)(netif_->node());	
	struct hdr_cmn* ch ;
	
	if (index_==0 && verbose_==1 ) printf("<%d> %f	 Up slot %d starts now. \n", 
			index_, NOW, up_slot_);
	
	// Restart timer for next slot.
	mhUpSlot_.start((Packet *)e, slot_time_);

	if (node_->base_stn() != node_->address()) {//node is an MS 
		
		if (pktTx[up_slot_] != NULL) {
 		
			//check if no more pkts to tx, start slot release timer
			if (q_ !=NULL && gprs_==1){
				len= q_->length();	
				//printf("<%d> %f MAC que=%0x length=%d \n", index_, NOW, q_, len);
			 	if  (up_slot_ != 0) {// tx_chan[0]=0 always.for random access/signals 
					if ( len==0 &&  (mhRel_.busy() ==0) && (tx_chan[up_slot_] !=-10) ) {
					//	slot_release_time_ =(2*(SLOTS_PER_FRAME-1) - up_slot_ +1)* slot_time_;
						slot_release_time_ = ((4*SLOTS_PER_FRAME) ) *slot_time_;
						mhRel_.start(slot_release_time_);
						if (verbose_==1)
						  printf("<%d> %f  release timer started \n",index_, NOW);
					}
				}	 
			} else {
		 	//printf("<%d> %f MAC que NULL q=%0x \n",index_, NOW, q_ ); 
			}
			
			//start transmission
			radioSwitch(ON);
		   
			ch =  HDR_CMN(pktTx[up_slot_]);
				  
			if (verbose_==1)
				    printf("<%d> %f in upslothandler. ptype=%s size=%d, slot=%d \n",
					index_, NOW, packet_info.name(ch->ptype()), ch->size(),up_slot_ );
			
			tx_onto_PHY ( pktTx[up_slot_] );
			pktTx[up_slot_] =0;
		}
	}
		
	if (node_->base_stn() == node_->address()) {//node is itself a BS 
	
		/* Do nothign here. taken care of in bs_recv() */

	}

	up_slot_ ++;
	if (up_slot_ == SLOTS_PER_FRAME	) {
		up_slot_ = 0; //ie wrap around.
	}
	return;
}

	
void MacGprs::recvHandler(Event *e) 
{
	u_int32_t dst, src; 
	int size;
	Packet * p= (Packet * )e;	
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_gprs *dh = HDR_MAC_GPRS(p);

	/* Check for collision.  */
	if  (ch->error() == EF_COLLISION) {
		ch->error() = 1;
	}	
	
	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);
	size = ch->size();
	
	if (verbose_==1)
	  printf("<%d> %f pkt [from %d to %d], size = %d uid=%d recv compleeted.\n", 
			index_, NOW, src, dst, size, ch->uid()); //richa

	radioSwitch(OFF);

	if ((dst != MAC_BROADCAST) && (dst != (u_int32_t)index_)) {
		// some error happened somewhere and the pkt landed up with me
		// somehow... so am dropping.
		drop(p);
		return;
	}
 
	fwd_DATA_to_LL(p);
	return;
}
 
// Called after transmission a packet ie when txTimer expires.
// Frees the pkt. Turns off the radio.
void MacGprs::sendHandler(Event *e) 
{
    if (verbose_==1) 
		printf("<%d> %f pkt uid =  %d send complete.\n", 
		 index_, NOW, HDR_CMN((Packet*)e)->uid());
	Packet::free((Packet *)e);
	radioSwitch(OFF);

	/* unlock IFQ if not waiting for res_reply. */
	if (mhwait_.busy() ==0 ) { 
		if(callback_) {
			Handler *h = callback_;
			callback_ = 0;
			h->handle((Event*) 0);
		} 
	}
	
	return;
}

// Called when the release timer expires
void MacGprs::releaseHandler()
{
	if (verbose_==1) printf("<%d> %f in release handler. \n" ,index_, NOW );
    send_let_go()  ;  
 	return;
}	 

// Called when the back off timer expires
void MacGprs::backoffHandler()
{  //resend resource request
	if (verbose_==1) printf("<%d> %f Backoff timer expired \n" , index_, NOW);
	send_res_request();
}


// Called when the wait timer expires
void MacGprs::waitHandler()
{ 
	int temp;
    double backoff_time;
	temp =  Random::integer(30);  
	backoff_time = ( double)( temp * slot_time_) ; 
	if (verbose_==1) 
		printf("<%d> %f res reply not recvd in 7 TS. will backoff for %fs (%d slots)\n", 
		index_, NOW, backoff_time, temp);
	// arbitly limited the backoff timer to 30 slots.- 
	// integer(K) gives a uniformly distributed integer btwn 0 and k-1 
	// maybe shud backoff for some num of frames instead.
	mhBackoff_.start(backoff_time); //start back off timer
}	

//create a resource release message
void MacGprs::send_let_go() 
{ // sent by an MS to its  BS
	
	int dst, j;
	Packet *p1 = Packet::alloc();
	
	struct hdr_cmn* ch1= HDR_CMN(p1);
	struct hdr_mac_gprs* dh1 = HDR_MAC_GPRS(p1);
	
	MobileNode * node_ = (MobileNode*)(netif_->node());
	//dst=node_->base_stn();	
	dst=0;
	
	this->hdr_src((char*)dh1, this->addr());
	this->hdr_dst((char*)dh1, dst);

	ch1->uid() = 0;
	ch1->direction() = hdr_cmn::DOWN;
	ch1->ptype() = PT_MAC;
    //ch1->size() = slot_packet_len_;
    ch1->size() = 20; //arbitly set at 20.
    ch1->iface() = -2;
	ch1->error() = 0; 
	
	if (verbose_==1)
	   printf("<%d> %f in send_let_go dst= %d\n", index_, NOW, /*node_->base_stn() */ dst);
	
	dh1->dh_fc.fc_type       = MAC_Type_Control;
	dh1->dh_fc.fc_subtype    = MAC_Subtype_tx_end;
	
	// find upslot/freq to tx on
	for ( j=1; j< SLOTS_PER_FRAME ; j++) {
		if ( tx_chan[j] != -10 ) {
				break; 
		 }
	 }
	
	ch1->chan() = tx_chan[j]; // sent on freq j, upslot 0.
	tx_chan[j]= -10; //release the channels...
	rx_chan[j]= -10;
	pktTx[j] = p1;
	return;
}

