/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */

/* By Sandeep Kumar, Kopparapu Suman and Richa Jain,
 * Indian Institute of Technology, Bombay.
 * June, 2001.
*/ 

/* Copyright (c) 2001 Indian Insitute of Technology, Bombay.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code and binary code must contain
 * the above copyright notice, this list of conditions and the following 
 * disclaimer.
 * 
 * 2. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: 
 * This product includes software developed at Indian Insitute of
 * Technology, Bombay. 
 *
 * 3. The name of the Institute may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission.
 * INDIAN INSTITUTE OF TECHNOLOGY, BOMBAY, MAKES NO REPRESENTATIONS 
 * CONCERNING EITHER THE MERCHANTABILITY OF THIS SOFTWARE OR THE 
 * SUITABILITY OF THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software 
 * is provided "as is" without express or implied warranty of any kind.
*/


#include <errmodel.h>
#include <mac.h>
#include <rlc.h>
#include <address.h>
#include <dsr/hdr_sr.h>
#include <mac-gprs.h>

int hdr_rlc::offset_; 

static class RLCHeaderClass : public PacketHeaderClass {
public:
	RLCHeaderClass(): PacketHeaderClass("PacketHeader/RLC",
					    sizeof(hdr_rlc)) {
		bind_offset(&hdr_rlc::offset_);
	}
} class_hdr_rlc;


static class RLCClass : public TclClass {
public:
	RLCClass() : TclClass("RLC") {}
	TclObject* create(int, const char*const*) {
		return (new RLC);
	}
} class_rlc;

int RLC::rlcverbose_ = {0};//rj

RLC::RLC() : LinkDelay(), seqno_(0), ackno_(0), macDA_(0), ifq_(0),
	mac_(0), lanrouter_(0), arptable_(0), varp_(0),
	downtarget_(0), uptarget_(0), lhSend_(this), acSend_(this),
	datacounter(0), ackcounter(0), window_(60), rackno_(-1), 
	acked_(0), rlcfraged_(1), rlcfragsz_(500),inseq_(1),unackseqno_(0),numdups(0)
{
	bind("macDA_", &macDA_);
	bind("acked_", &acked_);
	bind("rlcfraged_",&rlcfraged_);
	bind("rlcfragsz_",&rlcfragsz_);
    bind("rlcverbose_",&rlcverbose_); 
	
	buf_= new PacketQueue();
	Txbuf_ = new PacketQueue();
	Rxbuf_ = new PacketQueue();
}

int RLC::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 3) {
		if (strcmp(argv[1], "ifq") == 0) {
			ifq_ = (Queue*) TclObject::lookup(argv[2]);
			return (TCL_OK);
		}
		if(strcmp(argv[1], "arptable") == 0) {
                        arptable_ = (ARPTable*)TclObject::lookup(argv[2]);
                        assert(arptable_);
                        return TCL_OK;
                }
		if(strcmp(argv[1], "varp") == 0) {
                        varp_ = (VARPTable*)TclObject::lookup(argv[2]);
                        assert(varp_);
                        return TCL_OK;
                }
		if (strcmp(argv[1], "mac") == 0) {
			mac_ = (Mac*) TclObject::lookup(argv[2]);
                        assert(mac_);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "down-target") == 0) {
			downtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "up-target") == 0) {
			uptarget_ = (NsObject*) TclObject::lookup(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "lanrouter") == 0) {
			lanrouter_ = (LanRouter*) TclObject::lookup(argv[2]);
			return (TCL_OK);
		}

	}
	else if (argc == 2) {
		if (strcmp(argv[1], "ifq") == 0) {
			tcl.resultf("%s", ifq_->name());
			return (TCL_OK);
		}
		if (strcmp(argv[1], "mac") == 0) {
			tcl.resultf("%s", mac_->name());
			return (TCL_OK);
		}
		if (strcmp(argv[1], "down-target") == 0) {
			tcl.resultf("%s", downtarget_->name());
			return (TCL_OK);
		}
		if (strcmp(argv[1], "up-target") == 0) {
			tcl.resultf("%s", uptarget_->name());
			return (TCL_OK);
		}
	}
	return LinkDelay::command(argc, argv);
}



void RLC::recv(Packet* p, Handler* /*h*/)
{
	hdr_cmn *ch = HDR_CMN(p);
	hdr_rlc *rlch = HDR_RLC(p);//sandy
	hdr_ip *iph =HDR_IP(p);

	if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: recv pkt_type=%s uid=%d ipsrc=%d ipdst=%d\n",
			mac_->addr(), NOW, packet_info.name(ch->ptype()), ch->uid_,
			Address::instance().get_nodeaddr(iph->saddr()),
			Address::instance().get_nodeaddr(iph->daddr()));
// Sanity Check
	assert(initialized());

	if(p->incoming) {
                p->incoming = 0;
	}
//if arp packet send to approprite place  //sandy
	if(ch->direction()==hdr_cmn::UP){
			if(ch->ptype_ == PT_ARP){
				if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: sent an ARP going up to arptable\n", mac_->addr(), NOW);
				/* arptable_->arpinput(p, this); */
				Scheduler& s = Scheduler::instance();
				s.schedule(uptarget_, p, delay_);
				return;
			}
			if(ch->ptype_ == PT_LLACK) {
				if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: sent an llack going up \n", mac_->addr(), NOW);
				Scheduler& s = Scheduler::instance();
				s.schedule(uptarget_, p, delay_);
				return;
			}

	}
	
	if(ch->direction()==hdr_cmn::DOWN){
			if(ch->ptype_ == PT_ARP){
				if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: sent an ARP going down to arptable\n", mac_->addr(), NOW);
				Scheduler& s = Scheduler::instance();
				s.schedule(downtarget_, p, delay_);
				return;
			}
			if(ch->ptype_ == PT_LLACK){
				if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: sent an llack going down\n", mac_->addr(), NOW);
				Scheduler& s = Scheduler::instance();
				s.schedule(downtarget_, p, delay_);
				return;
			}

	}


//if ack pkt or data pkt...  //sandy
	if(rlch->rlctype_==RLC_ACK)
		recvACK(p);
	else
		recvDATA(p);

}


void RLC::recvDATA(Packet* p)
{
	hdr_cmn *ch = HDR_CMN(p);
	
	if(ch->direction() == hdr_cmn::UP){
		if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: DATA on its way up. call sendupDATA\n", mac_->addr(), NOW);	
		sendUpDATA(p);
	}else{
		ch->direction() = hdr_cmn::DOWN;
		if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: DATA going down. call enqueDATA\n", mac_->addr(), NOW);
		enqueDATA(p);
	}
}


void RLC::enqueDATA(Packet* p)
{
	hdr_cmn *ch = HDR_CMN(p);
	hdr_ip *ih = HDR_IP(p);

	hdr_rlc *rlch = HDR_RLC(p);
	
	rlch->rlctype() = RLC_DATA;
	

//fragmenting to start here....//sandy
	int psize;
	if(rlcfraged_==1)
		psize=rlcfragsz_;
	else
		psize=ch->size_;

		int bopno=seqno_;
		int eopno=bopno + (int)((ch->size_-1)/psize);  
		
		if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: eopno=%d, bopno=%d \n", mac_->addr(), NOW, eopno, bopno);
				 
		int i=seqno_;
		
		for ( i = bopno; i <= eopno; i++){
			Packet * sp = p->copy();
			hdr_cmn *chsp = HDR_CMN(sp);
			hdr_rlc *rlchsp = HDR_RLC(sp);
		
			rlchsp->seqno_ = 0;
			rlchsp->bopno_ = bopno;
			rlchsp->eopno_ = eopno;
			rlchsp->psize_ = chsp->size_; //hack for now to remeber uid 
			chsp->size_  = psize;
			chsp->uid_= i; //rj
		//	chsp->next_hop_ = -1; //was a debug option for non-ack mode

			
			if((acked_==0)||(chsp->next_hop()==-1)){
				rlchsp->seqno_=i;
				Scheduler& s = Scheduler::instance();
				s.schedule(downtarget_, sp, delay_);
				if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: sending NON-acked down. pkt uid=%d\n",
						mac_->addr(), NOW,chsp->uid_);
			}else{ 
				rlchsp->seqno_ =i;
				seqno_++;
				buf_->enque(sp);    
				if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: buffering frag. rlc_seqno=%d for pkt uid=%d buf_len=%d\n"
						, mac_->addr(), NOW, rlchsp->seqno_,chsp->uid_, buf_->length());
			}
		}	
		if((acked_==1))
			sendDownDATA();
}

void RLC::sendDownDATA(void)
{
	Packet* sp;
	if(!(sp = buf_->deque())){
		if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: que is empty. length=%d\n", mac_->addr(), NOW,buf_->length());
		return;
	}
	
	hdr_rlc *rlcsp=HDR_RLC(sp);
	
	if(Txbuf_->length()<window_){
		Txbuf_->enque(sp);
		pktTx_ = sp->copy();	
		Scheduler& s = Scheduler::instance();
		s.schedule(downtarget_, sp->copy(), delay_);
		lhSend_.stop();
		lhSend_.start(0.4);
		if (rlcverbose_==1) fprintf (stderr, "<%d> %f RLC: Txbuf not full. sending a data pkt down, rlc_seqno=%d from buffer \n", 
				mac_->addr(), NOW, rlcsp->seqno_);
		sendDownDATA();
	}else{
		buf_->enqueHead(sp);
		if (rlcverbose_==1) fprintf (stderr, "<%d> %f RLC: Txbuf full. couldnt send rlc_seqno=%d \n",
				mac_->addr(), NOW, rlcsp->seqno_);
	}
}


void RLC::recvACK(Packet* p)
{
	hdr_rlc *rlch = HDR_RLC(p);

	//rj
	if (rlch->ackno_==ackno_) {
		if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: recvd duplicate RLC_ACK num %d\n",
				            mac_->addr(), NOW, rlch->ackno_);      
		sendDownDATAonACK(); 
		return;
	}	
	else if(ackno_ < rlch->ackno_){
		ackno_ = rlch->ackno_;
		lhSend_.stop();
		datacounter=0;
	}
	
	if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: recvd RLC_ACK num %d\n",
			mac_->addr(), NOW, rlch->ackno_);

	Packet* sp=Txbuf_->head();
	int Txlen=Txbuf_->length();
	
	//delete pkts with seqno < ackno
	for (int i=1; i<Txlen; i++){
		hdr_rlc *rlchsp= HDR_RLC(sp);
		if((rlchsp->seqno_)<ackno_){
			Txbuf_->remove(sp);
			if (rlcverbose_==1) fprintf(stderr, "<%d> %f RLC: removed ie deleted pkt from tx buffer; rlc seqno=%d\n",
					mac_->addr(), NOW, rlchsp->seqno_);
		}
		sp = sp->next_;
	}
	
	// I dont want to send down apkt when i get an ack. rj
	if(Txbuf_->length()==window_){
		//if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: txbuf full, calling sendDownDATAonACK\n", mac_->addr(), NOW);
	//	sendDownDATAonACK();
	}
	else{
		if(buf_->length()==0){
		//	if (rlcverbose_==1) fprintf (stderr, "<%d> %f RLC: no more new pkts to send, calling sendDownDATAonACK\n", 
		//			mac_->addr(), NOW);
	//		sendDownDATAonACK();
		}
		else{
			if (rlcverbose_==1) fprintf (stderr, "<%d> %f RLC: got an ack. calling sendDOWNDATA\n" , mac_->addr(), NOW); 
			sendDownDATA(); //hmm, this not called at all rj
		}	
	}

}


void RLC::sendDownDATAonACK()
{
	Packet* sp=Txbuf_->head();
	int Txlen_=Txbuf_->length();

	for(int i=1; i<Txlen_; i++){
		
		hdr_rlc *rlchsp= HDR_RLC(sp);
		
		if((rlchsp->seqno_)==ackno_){
			pktTx_ = sp->copy();	
			if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: retx pkt from txbuf seqno=%d, eopno=%d\n",
					mac_->addr(), NOW,rlchsp->seqno_,rlchsp->eopno_);
			Scheduler& s = Scheduler::instance();
			s.schedule(downtarget_, sp->copy(), delay_);
				lhSend_.stop();
				lhSend_.start(0.4);
			return;
		}
		sp = sp->next_;
	}
		
}

void RLC::sendUpDATA(Packet* p)
{
	acSend_.stop();
	ackcounter=0;
	if (hdr_cmn::access(p)->error() > 0)
		drop(p);
	else{
		hdr_cmn *ch = HDR_CMN(p);
		hdr_rlc *rlch = HDR_RLC(p);
		hdr_ll *llch = HDR_LL(p);
		hdr_ip *iph = HDR_IP(p);
	
		if((acked_==0)||(ch->next_hop()==-1)){
           if (rlch->seqno_==unackseqno_) 
                unackseqno_++; 
            else{ 
                inseq_=0;
                unackseqno_=rlch->seqno_;
 			}				
			
			if (rlch->seqno_ == rlch->eopno_ ){
				if(inseq_){ 
		//			ch->uid_=rlch->psize_; //sandy;
		  			if (rlcverbose_==1) fprintf(stderr ,"<%d> %f RLC: nonack. recvd all rlc frags for pktuid=%d ll_seqno=%d \n",
							mac_->addr(), NOW, ch->uid_, llch->seqno_);   
					Scheduler& s = Scheduler::instance();
					s.schedule(uptarget_, p, delay_);
				}
				else {
					if (rlcverbose_==1) fprintf(stderr ,"<%d> %f RLC: nonack. pkts not inseq. not sent up\n",mac_->addr(), NOW);
                }

			    inseq_=1;
				unackseqno_=0; 
			} else { 
			   if (rlcverbose_==1) fprintf(stderr ,"<%d> %f RLC: nonack. recvd rlc_seqno=%d for pkt uid=%d \n"
					   , mac_->addr(), NOW, rlch->seqno_,  ch->uid_ );   
				  Packet::free(p);   
			}
			return;
		}else{
		
		if(rlch->seqno_==(rackno_+1)){
			rackno_++;
			numdups=0;

		//	ch->uid_=rlch->psize_; //sandy;
			if (rlcverbose_==1) fprintf(stderr , "<%d> %f RLC: rackno=%d, src=%d \n",  mac_->addr(), NOW,rackno_,
					 Address::instance().get_nodeaddr(iph->saddr()) );//rj
			if ( rlch->seqno_ == rlch->eopno_ ){
				ch->size_=rlch->psize_;
		  		if (rlcverbose_==1) fprintf(stderr , "<%d> %f RLC:ACK:. send up to LL pktuid=%d llseqno=%d \n",
						mac_->addr(), NOW,ch->uid_, llch->seqno_);   
				Scheduler& s = Scheduler::instance();
				s.schedule(uptarget_, p, delay_);
			} else { 
			   if (rlcverbose_==1) fprintf(stderr ,"<%d> %f RLC:ACK: recvd rlc_seqno=%d for pkt uid=%d ll_seqno=%d \n"
					   , mac_->addr(), NOW, rlch->seqno_,  ch->uid_ , llch->seqno_);   
				 // Packet::free(p);   
			}
//			int choose=1;	
			int flag=1;
//			Packet * spcpy;
			while(flag){
				flag=0;
				if(Rxbuf_->length()!=0){
					Packet* sp=Rxbuf_->head();
					int Rxlen_=Rxbuf_->length();
					for(int i=1;i<=Rxlen_;i++){
						hdr_rlc* rlcsp = HDR_RLC(sp);
						if(rlcsp->seqno_==(rackno_+1)){
							rackno_++;
							flag=1;
//							choose=0;
//							Packet * spcpy=sp->copy();
							hdr_cmn* chsp=HDR_CMN(sp);
							if ( rlcsp->seqno_ == rlcsp->eopno_ ){//??rj
								chsp->size_=rlcsp->psize_;
		  						if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: pkt in rxbuf: send up uid=%d seqno=%d bopno=%d eopno=%d \n",
										mac_->addr(), NOW,ch->uid_,rlch->seqno_,rlch->bopno_,rlch->eopno_);   
								Scheduler& s = Scheduler::instance();
								s.schedule(uptarget_, sp->copy(), delay_);
							} else { 
				  				 if (rlcverbose_==1) fprintf(stderr , "<%d> %f RLC: pkt in rxbuf: drop uid=%d seqno=%d bopno=%d eopno=%d next_hop %d \n", mac_->addr(), NOW,chsp->uid_,rlcsp->seqno_,rlcsp->bopno_,rlcsp->eopno_,chsp->next_hop());   
							}
							Rxbuf_->remove(sp);
					}
					sp=sp->next_;
				}
			}
			}
//			if (choose)
				sendACK(p);
//			else
//				sendACK(spcpy);

	}else if(rlch->seqno_>(rackno_+1)){
		Rxbuf_->enque(p);
	     if (numdups<1){
			if (rlcverbose_==1) fprintf (stderr,"<%d> %f RLC: pkt out of seq, buf in rxbuf;send duplicate ackseqno=%d\n", mac_->addr(), NOW,rlch->seqno_);
			sendACK(p); //rj
			numdups ++;
		}
	}
		
	}
	}
}

void RLC::sendACK(Packet* p)
{
		hdr_cmn *ch = HDR_CMN(p);
//		hdr_rlc *rlch = HDR_RLC(p);
		hdr_ip *iph = HDR_IP(p);
		hdr_mac_gprs *dh =HDR_MAC_GPRS(p);
//		char *mh = (char*)p->access(hdr_mac::offset_);
	
		Packet* sp = Packet::alloc();
		
		hdr_cmn *chsp = HDR_CMN(sp);
		hdr_rlc *rlchsp = HDR_RLC(sp);
		hdr_ip *iphsp = HDR_IP(sp);
		char *mhsp = (char *)HDR_MAC(sp);
// 		char *mhsp = (char*)sp->access(hdr_mac::offset_);
		
//		iphsp->src_ = iph->dst_;
//		iphsp->dst_ = iph->src_;

//		chsp->next_hop_ = ch->last_hop_;
//		chsp->last_hop_ = ch->next_hop_;
		chsp->next_hop_=Address::instance().get_nodeaddr(iph->saddr());	
		chsp->prev_hop_=Address::instance().get_nodeaddr(iph->daddr());	

		chsp->addr_type_ = ch->addr_type_;
		chsp->direction_ = hdr_cmn::DOWN;
		chsp->ptype_=PT_RLCACK; 
		chsp->size_=6;
		chsp->iface()= -2;
		chsp->error()= 0;
		chsp->uid()= rackno_+1;
		
		u_int32_t dst_=ETHER_ADDR(dh->dh_sa);
		mac_->hdr_src(mhsp, mac_->addr());
		mac_->hdr_type(mhsp, ETHERTYPE_IP);
		mac_->hdr_dst(mhsp, dst_);
//		mac_->hdr_dst(mhsp, chsp->next_hop());
		
		rlchsp->rlctype() = RLC_ACK;
		rlchsp->ackno_ = rackno_+1;
	
		pktRx_ = sp->copy();	
		Scheduler& s = Scheduler::instance();
		s.schedule(downtarget_, sp, delay_);
		acSend_.stop();
		acSend_.start(0.4);
		if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: send RLCACK num %d to mac_dst=%d\n", 
				mac_->addr(), NOW, rlchsp->ackno_, dst_);
		
//		Packet::free(p);   

}

void RLC::sendHandler()
{
	send_timer();
}

void RLC::sendackHandler()
{
	send_acktimer();
}


void RLC::send_timer()
{
	RetransmitDATA();
//	RetransmitACK();
}

void RLC::send_acktimer()
{
	RetransmitACK();
}


void RLC::RetransmitDATA()
{
	datacounter =datacounter + 1;
	assert(pktTx_);
 	hdr_rlc* rlch=HDR_RLC(pktTx_);
	
	if (datacounter <3){	
	if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: retx rlc frag seqno=%d bopno=%d eopno=%d for %dth time\n", 
			mac_->addr(), NOW,rlch->seqno_,rlch->bopno_,rlch->eopno_,datacounter);
		Scheduler& s = Scheduler::instance();
		s.schedule(downtarget_, pktTx_->copy(), delay_);
		lhSend_.stop();
		lhSend_.start(0.4);
	}
	else{
	if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: retxed rlc frag 5 times. giveup. rlc seqno=%d bopno=%d eopno=%d\n",
			mac_->addr(), NOW,rlch->seqno_,rlch->bopno_,rlch->eopno_);
		//datacounter=0;
		ackno_++;
		sendDownDATA();
	}
}

void RLC::RetransmitACK()
{
	ackcounter =ackcounter +1;
	assert(pktRx_);
	hdr_rlc *rlch=HDR_RLC(pktRx_);
	if (ackcounter < 3){
		if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: retransmit rlc ack %d for %dth time\n",
				mac_->addr(), NOW,rlch->ackno_,ackcounter);
		Scheduler& s = Scheduler::instance();
		s.schedule(downtarget_,pktRx_->copy(),delay_);
		acSend_.stop();
		acSend_.start(0.4);
	}
	else{
		if (rlcverbose_==1) fprintf(stderr,"<%d> %f RLC: transmitted rlc ack num %d 3 times giveup\n", 
				mac_->addr(), NOW, rlch->ackno_);
//		rackno_++;
		ackcounter=0;
	}
}
